# Copyright (c) 2016 Oracle and/or its affiliates. All rights reserved.
# created by sahit.gollapudi

__author__ = "sahit"

import os
import sys
import json
import uuid
import shutil
import logging 
import requests
from collections import OrderedDict
try:
    from executor import RequestExecutor
    from utils import Utils
    from __init__ import __version__
    from opcservice import OPCService
    from messages import Messages, ErrorMessages
    from exceptions import PSMUsageError, OSSUploadError, OSSDetailsError
    from internal.opc_cloud_storage_handler import OPCCloudStorageHandler
except:
    from .executor import RequestExecutor
    from .utils import Utils
    from .__init__ import __version__
    from .opcservice import OPCService
    from .messages import Messages, ErrorMessages
    from .exceptions import PSMUsageError, OSSUploadError, OSSDetailsError
    from .internal.opc_cloud_storage_handler import OPCCloudStorageHandler

class CustomAccsPushExecutor(RequestExecutor):
    '''
        Custom Impl for Accs Push.
    '''
    def __init__(self, reqParamHandler, usage, prog):
        super().__init__(reqParamHandler)
        self.usage = usage 
        self.prog = prog
        self.max_app_size = 5368709122 # 5 GB #262144000 #250 MB
        self.max_limit_accs = 1048576000 #1000 MB
        self.max_limit_archive_warning = 262144000 # 250 MB
        self._archive_path = 'archivePath'
        self._archive_url = 'archiveURL'
        self.archive_args_list = [self._archive_url, self._archive_path]
    
    def pre_execute_request(self):
        #print(self.reqParamHandler.cmd_params_values)
        # validate for archivePath and archiveURL. Only one should be 
        # specifed either archivePath or archiveUrl
        if all(elem in self.reqParamHandler.cmd_params_values for elem in self.archive_args_list):
            archive_args = map(lambda x: "-{0}/{1}".format(self.reqParamHandler.arg_parameter_list[x].alias_name, \
                                                              self.reqParamHandler.arg_parameter_list[x].argument_option_name), \
                                                              self.archive_args_list)
            err_msg = ErrorMessages.OPAAS_CLI_SPECIFY_ONE_ARGUMENT_ERROR.format(" or ".join(list(archive_args)))
            raise PSMUsageError(usage = "usage: %s" % self.usage, prog = self.prog, err_msg = err_msg)
        
        if self._archive_path in self.reqParamHandler.cmd_params_values:  
            isArchivePathDir = False          
            # 1) Check if archive path is a directory to be zipped 
            # or it is a zip file. 
            archive_path = self.reqParamHandler.cmd_params_values[self._archive_path]
            archive_path_arg = "-{0}/{1}".format(self.reqParamHandler.arg_parameter_list[self._archive_path].alias_name, \
                                                self.reqParamHandler.arg_parameter_list[self._archive_path].argument_option_name)
            # raise an exception if the path doesnt exist.
            if not os.path.exists(archive_path):
                err_msg = ErrorMessages.OPAAS_CLI_NO_SUCH_FILE_EXISTS_ERROR.format(archive_path_arg, archive_path)
                raise PSMUsageError(usage = "usage: %s" % self.usage, prog = self.prog, err_msg = err_msg)
            
            # Determine the storage URL, container, auth mechanism and max_upload_size
            response_json, max_upload_size = self._get_cloud_storage_details()
            # convert the max_upload size to bytes
            if max_upload_size is not None:            
                self.max_limit_accs = self._utils.convert_size_to_bytes(int(max_upload_size))
            
            # Check if the given path is a dir.
            if os.path.isdir(archive_path):
                # create a zip in the temp dir: TODO
                isArchivePathDir = True
                if archive_path.endswith('/'):
                    archive_path = archive_path[:-1]
                temp_name = self._utils._tmp + os.sep + os.path.basename(archive_path)
                zip_extension = "zip"
                archive_path = shutil.make_archive(temp_name, zip_extension, archive_path)
            
            # 2) check if the zipped file size is less than equal to 5 GB.
            archive_path_size = os.stat(archive_path).st_size
            if archive_path_size > self.max_limit_accs:
                err_msg = ErrorMessages.OPAAS_CLI_MAX_LIMIT_ERROR.format(archive_path_arg, self._utils.get_human_readable_size(self.max_limit_accs), self.reqParamHandler.cmd_params_values['archivePath'])
                raise PSMUsageError(usage = "usage: %s" % self.usage, prog = self.prog, err_msg = err_msg)
            
            # display warning msg if size is greater than 250 MB
            if archive_path_size > self.max_limit_archive_warning:
                sys.stdout.write(Messages.OPAAS_CLI_ACCS_PUSH_SIZE_GREATER_THAN_LIMIT_WARNING % (self._utils.get_human_readable_size(archive_path_size), \
                                                                                                 self._utils.get_human_readable_size(self.max_limit_archive_warning)))
            # display a msg to the user as it takes time to execute 
            sys.stdout.write(Messages.OPAAS_CLI_LONG_CMD_EXEC_INFO_MSG)
            # 3) Upload the file to the cloud storage and compute Archive URL TODO
            try:
                archive_url = self._upload_app_to_cloud(archive_path, response_json) 

                # 4) on success, store the archiveUrl into the requestHandler object TODO
                self.reqParamHandler.cmd_params_values[self._archive_url] = archive_url
                # remove the archivePath from the dict.
                del self.reqParamHandler.cmd_params_values[self._archive_path]
            
                # Log the command to be executed after custom changes
                self.reqParamHandler.log_command_executed()
            finally:
                if isArchivePathDir:
                    os.remove(temp_name + "." + zip_extension)                 

        # 5) call the main execute_request which is commented below. 
        self.execute_request()
    
    def _upload_app_to_cloud(self, archive_path, response_json):
        
        # TODO: remove these lines
        # Determine the storage URL, container and auth mechanism
        # response_json = self._get_cloud_storage_details()
        
        # build opc cloud json args to upload the archive.              
        opc_cloud_storage_json = OrderedDict()
        try:
            # TODO: Down the line the upload limit for accs push will be returned as a header param.
            cloud_container = response_json['container'] #"ImportExportPoc"
            opc_cloud_storage_json['cloud_storage_url'] = response_json['storageURL'] #"https://storage.oraclecorp.com/v1/Storage-StorageEval01admin"
            opc_cloud_storage_json['cloud_auth_token'] = response_json['authToken'] #None
            opc_cloud_storage_json['delete'] = True
            #opc_cloud_storage_json['cloud_user'] = "StorageEval01admin.Storageadmin"
            #opc_cloud_storage_json['cloud_pwd'] = "Welcome1"
            
            # the format for archiveUrl is container/uuid/filename. use the '/' for consistency
            archive_url = "/".join([cloud_container, str(uuid.uuid4()), os.path.basename(archive_path)])        
            opc_cloud_storage_json['source'] = archive_path
            opc_cloud_storage_json['archive_url'] = archive_url
            opc_cloud_storage_handler = OPCCloudStorageHandler()
            upload_result = opc_cloud_storage_handler._upload_archive(opc_cloud_storage_json)
            
            if upload_result.get_status_code() == 0:
                # upload was successful
                logger.info("{0}: {1}".format(upload_result.get_http_code(), upload_result.get_message()))
                return archive_url 
            else:
                logger.error("{0}: {1}".format(upload_result.get_http_code(), upload_result.get_message()))
                raise OSSUploadError(http_code = upload_result.get_http_code(), err_msg = upload_result.get_message())
        except KeyError as e:
            logger.error(str(e))
            raise OSSDetailsError()
            
    
    def _get_cloud_storage_details(self):
        # Headers
        headers = OrderedDict()
        identityDomain = self._utils.getValueFromConfigFile(self._utils.identity_domain)
        headers[self._utils.authorization_header_key] = self._utils.getAuthToken()
        headers[self._utils.identity_domain_header_key] = identityDomain   
        headers[self._utils.cli_request_key] = 'cli' # Value doesn't matter
        headers[self._utils.cli_header_version_key] = __version__
        
        opc_service = OPCService()
        # Get cloud Storage details
        request_url = self._utils.getValueFromConfigFile(self._utils.default_uri) + \
                        self._utils._cloud_storage_endpoint.replace('{' + self._utils.identity_domain + '}', identityDomain)
        response_json, max_upload_size = opc_service.get_details(method='GET',request_url=request_url, headers=headers)
        
        return response_json, max_upload_size
        
#=================CONSTANTS==================#
# Custom impl Class Mapping:
CUSTOM_IMPL_MAP = {
                    'push' : CustomAccsPushExecutor
} 

logger = logging.getLogger(__name__)
